@@ -1,5 +1,3 @@
-(* $Id$ *)
-
 (*
   A fairly complete demonstration of the features provided
   by Easy-format.
@@ -9,7 +7,7 @@
 open Easy_format
 
 
-let list = 
+let list =
   { list with
       list_style = Some "list";
       opening_style = Some "op";
@@ -22,14 +20,14 @@ let label = { label with label_style = Some "label" }
 
 
 
-let tuple_param = 
+let tuple_param =
   { list with
       space_after_opening = false;
       space_before_closing = false;
       align_closing = false
   }
 
-let operator_param = 
+let operator_param =
   { list with
       space_after_opening = false;
       space_before_closing = false;
@@ -74,19 +72,19 @@ let format_float x =
   Atom (Printf.sprintf "%.5f" x, atom)
 
 let format_sum ?(wrap = `Wrap_atoms) l =
-  List (("(", "+", ")", { operator_param with wrap_body = wrap }), 
+  List (("(", "+", ")", { operator_param with wrap_body = wrap }),
 	List.map format_int l)
 
 let format_array ~align_closing ~wrap f a =
   let l = Array.to_list (Array.map f a) in
-  List (("[|", ";", "|]", 
+  List (("[|", ";", "|]",
 	 { list with
 	     align_closing = align_closing;
 	     wrap_body = wrap }),
 	l)
 
-let format_matrix 
-    ?(align_closing1 = true) 
+let format_matrix
+    ?(align_closing1 = true)
     ?(align_closing2 = true)
     ?(wrap1 = `Wrap_atoms)
     ?(wrap2 = `Wrap_atoms)
@@ -96,15 +94,15 @@ let format_matrix
 
 
 let format_record f l0 =
-  let l = 
-    List.map 
-      (fun (s, x) -> Label ((Atom (s ^ ":", atom), label), f x)) 
+  let l =
+    List.map
+      (fun (s, x) -> Label ((Atom (s ^ ":", atom), label), f x))
       l0 in
   List (("{", ";", "}", list), l)
 
-let begin_style = 
+let begin_style =
   { label with indent_after_label = 0 },
-  ("begin", ";", "end", 
+  ("begin", ";", "end",
    { list with stick_to_label = false })
 
 let curly_style =
@@ -116,14 +114,66 @@ let format_function_definition (body_label, body_param) name param body =
     (
       Label (
 	(Atom ("function " ^ name, atom), label),
-	List (("(", ",", ")", tuple_param), 
+	List (("(", ",", ")", tuple_param),
 	      List.map (fun s -> Atom (s, atom)) param)
-      ), 
+      ),
       body_label
     ),
     List (body_param, List.map (fun s -> Atom (s, atom)) body)
   )
 
+(*
+   Illustrate the difference between `Force_break and `Force_breaks_rec on
+   labels.
+*)
+let label_one_atom = Atom ("reallyLongLabelOne", atom)
+let label_two_atom = Atom ("reallyLongLabelTwo", atom)
+let label_three_atom = Atom ("reallyLongLabelABC", atom)
+let make_list_in_labels (wrap) =
+  Label (
+    (label_one_atom, label),
+    (
+      Label (
+        (label_two_atom, label),
+        (
+          Label (
+            (label_three_atom, label),
+            List (
+              ("[", ",", "]", { list with wrap_body = wrap }),
+              [
+                Atom ("1.23456", atom);
+                Atom ("9.87654", atom);
+                Atom ("9.87654", atom);
+                Atom ("9.87654", atom);
+              ]
+            )
+          )
+        )
+      )
+    )
+  )
+
+(*
+   Illustrate the difference between `Force_break and `Force_breaks_rec
+*)
+let make_heterogenous_list (container_wrap, wrap) =
+  List (
+    ("[", ",", "]", { list with wrap_body = container_wrap }),
+    [
+      Atom ("0", atom);
+      List (
+        ("[", ",", "]", { list with wrap_body = wrap }),
+        [
+          Atom ("1.23456", atom);
+          Atom ("9.87654", atom);
+        ]
+      );
+      Atom ("1", atom);
+      Atom ("2", atom);
+      Atom ("3", atom);
+    ]
+  )
+
 let print_margin fmt () =
   let margin = Format.pp_get_margin fmt () in
   for i = 1 to margin do
@@ -148,17 +198,23 @@ let print s =
 let print_tuple fmt l =
   Pretty.to_formatter fmt (format_tuple format_int l)
 
+let print_heterogenous_list fmt wrap =
+  Pretty.to_formatter fmt (make_heterogenous_list wrap)
+
+let print_list_in_labels fmt wrap =
+  Pretty.to_formatter fmt (make_list_in_labels wrap)
+
 let print_sum ?wrap fmt l =
   Pretty.to_formatter fmt (format_sum ?wrap l)
 
 let print_matrix ?align_closing1 ?align_closing2 ?wrap1 ?wrap2 m fmt () =
-  Pretty.to_formatter fmt 
+  Pretty.to_formatter fmt
     (format_matrix ?align_closing1 ?align_closing2 ?wrap1 ?wrap2 m)
 
 let print_function_definition style name param fmt body =
   Pretty.to_formatter fmt (format_function_definition style name param body)
 
-let _ =
+let () =
   let ints = Array.to_list (Array.init 10 (fun i -> i)) in
 
   (* A simple tuple that fits on one line *)
@@ -171,7 +227,45 @@ let _ =
   with_margin 20 (print_sum ~wrap:`Always_wrap) ints;
   with_margin 20 (print_sum ~wrap:`Never_wrap) ints;
 
+  (* Heterogenous list *)
+  print "wrappable outer list, inner list using `Force_breaks";
+  with_margin 80 print_heterogenous_list (`Always_wrap, `Force_breaks);
+  with_margin 20 print_heterogenous_list (`Always_wrap, `Force_breaks);
 
+  print "wrappable outer list, inner list using `Force_breaks_rec";
+  with_margin 80 print_heterogenous_list (`Always_wrap, `Force_breaks_rec);
+  with_margin 20 print_heterogenous_list (`Always_wrap, `Force_breaks_rec);
+
+  print "never_wrap outer list, inner list using `Force_breaks";
+  with_margin 80 print_heterogenous_list (`Never_wrap, `Force_breaks);
+  with_margin 20 print_heterogenous_list (`Never_wrap, `Force_breaks);
+
+  print "never_wrap outer list, inner list using `Force_breaks_rec";
+  with_margin 80 print_heterogenous_list (`Never_wrap, `Force_breaks_rec);
+  with_margin 20 print_heterogenous_list (`Never_wrap, `Force_breaks_rec);
+
+  print "no breaks outer list, inner list using `Force_breaks";
+  with_margin 80 print_heterogenous_list (`No_breaks, `Force_breaks);
+  with_margin 20 print_heterogenous_list (`No_breaks, `Force_breaks);
+
+  print "no breaks outer list, inner list using `Force_breaks_rec";
+  with_margin 80 print_heterogenous_list (`No_breaks, `Force_breaks_rec);
+  with_margin 20 print_heterogenous_list (`No_breaks, `Force_breaks_rec);
+
+  print "label with inner list using `Force_breaks_rec";
+  with_margin 80 print_list_in_labels (`Force_breaks_rec);
+  with_margin 70 print_list_in_labels (`Force_breaks_rec);
+  with_margin 20 print_list_in_labels (`Force_breaks_rec);
+
+  print "label with inner list using `Force_breaks";
+  with_margin 80 print_list_in_labels (`Force_breaks);
+  with_margin 70 print_list_in_labels (`Force_breaks);
+  with_margin 20 print_list_in_labels (`Force_breaks);
+
+  print "label with inner list using `Never_wrap";
+  with_margin 80 print_list_in_labels (`Never_wrap);
+  with_margin 70 print_list_in_labels (`Never_wrap);
+  with_margin 20 print_list_in_labels (`Never_wrap);
 
   (* Triangular array of arrays showing wrapping of lists of atoms *)
   let m = Array.init 20 (fun i -> Array.init i (fun i -> sqrt (float i))) in
@@ -182,22 +276,22 @@ let _ =
 
   (* Other styles *)
   print "style 1";
-  with_margin 80 (print_matrix 
+  with_margin 80 (print_matrix
 		    ~align_closing1: false ~align_closing2: false m) ();
   print "style 2";
   with_margin 80 (print_matrix
 		    ~align_closing1: false ~align_closing2: false
 		    ~wrap2: `Never_wrap m) ();
   print "style 3";
-  with_margin 80 (print_matrix 
+  with_margin 80 (print_matrix
 		    ~align_closing1: false ~align_closing2: false
 		    ~wrap2: `Always_wrap m) ();
   print "style 4";
-  with_margin 80 (print_matrix 
+  with_margin 80 (print_matrix
 		    ~align_closing2: false
 		    ~wrap1: `Always_wrap ~wrap2: `Always_wrap m) ();
   print "style 5";
-  with_margin 80 (print_matrix 
+  with_margin 80 (print_matrix
 		    ~align_closing1: false
 		    ~wrap1: `Always_wrap ~wrap2: `Always_wrap m) ();
   print "style 6";
@@ -220,7 +314,7 @@ let _ =
     with_margin ~html margin
       (print_function_definition
 	 style
-	 "hello" ["arg1";"arg2";"arg3"]) 
+	 "hello" ["arg1";"arg2";"arg3"])
       [
 	"print \"hello\"";
 	"return (1 < 2)"
@@ -234,4 +328,3 @@ let _ =
 	  program true margin style
       ) [ 10; 20; 30; 40; 80 ]
   ) [ curly_style; begin_style ]
-
